// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright (c) 2015  Giulio Camuffo <giuliocamuffo@gmail.com>
// gamma-wl.h -- Wayland gamma adjustment header

#ifndef REDSHIFT_GAMMA_WAYLAND_H
#define REDSHIFT_GAMMA_WAYLAND_H

#include <stdint.h>

#include <wayland-client.h>

#include "redshift.h"

extern const gamma_method_t wl_gamma_method;


#endif /* ! REDSHIFT_GAMMA_DRM_H */
