// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright (c) 2014-2017  Jon Lund Steffensen <jonlst@gmail.com>
// location-geoclue2.h -- Geoclue2 location provider header

#ifndef REDSHIFT_LOCATION_GEOCLUE2_H
#define REDSHIFT_LOCATION_GEOCLUE2_H

#include "redshift.h"

extern const location_provider_t geoclue2_location_provider;

#endif /* ! REDSHIFT_LOCATION_GEOCLUE2_H */
