// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright (c) 2014  Mattias Andrée <maandree@member.fsf.org>
// Copyright (c) 2017  Jon Lund Steffensen <jonlst@gmail.com>
// gamma-drm.h -- DRM gamma adjustment header

#ifndef REDSHIFT_GAMMA_DRM_H
#define REDSHIFT_GAMMA_DRM_H

#include "redshift.h"

extern const gamma_method_t drm_gamma_method;

#endif /* ! REDSHIFT_GAMMA_DRM_H */
