// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright (c) 2010-2017  Jon Lund Steffensen <jonlst@gmail.com>
// gamma-randr.h -- X RANDR gamma adjustment header

#ifndef REDSHIFT_GAMMA_RANDR_H
#define REDSHIFT_GAMMA_RANDR_H

#include "redshift.h"

extern const gamma_method_t randr_gamma_method;

#endif /* ! REDSHIFT_GAMMA_RANDR_H */
