// SPDX-License-Identifier: GPL-3.0-or-later
// Copyright (c) 2010-2017  Jon Lund Steffensen <jonlst@gmail.com>
// location-manual.h -- Manual location provider header

#ifndef REDSHIFT_LOCATION_MANUAL_H
#define REDSHIFT_LOCATION_MANUAL_H

#include "redshift.h"

extern const location_provider_t manual_location_provider;

#endif /* ! REDSHIFT_LOCATION_MANUAL_H */
